package org.cobbzilla.util.io;

import lombok.extern.slf4j.Slf4j;
import org.cobbzilla.util.system.Bytes;

import java.io.*;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;

import static org.cobbzilla.util.daemon.ZillaRuntime.*;
import static org.cobbzilla.util.reflect.ReflectionUtil.closeQuietly;
import static org.cobbzilla.util.reflect.ReflectionUtil.instantiate;

@Slf4j
public class FilterInputStreamViaOutputStream extends PipedInputStream implements Runnable {

    private static final int DEFAULT_PIPE_BUFFER_SIZE = (int) (64 * Bytes.KB);
    private static final int DEFAULT_COPY_BUFFER_SIZE = (int) (8 * Bytes.KB);
    private static final long THREAD_TERMINATE_TIMEOUT = TimeUnit.SECONDS.toMillis(10);

    private final InputStream in;
    private final String name;
    private final PipedOutputStream pipeOut;
    private final OutputStream out;
    private final Thread thread;

    public Class<? extends OutputStream> getOutputStreamClass () { return out == null ? null : out.getClass(); }

    public FilterInputStreamViaOutputStream(InputStream in, Class<? extends OutputStream> outStreamClass) {
        this(in, outStreamClass, null);
    }

    public FilterInputStreamViaOutputStream(InputStream in, Class<? extends OutputStream> outStreamClass, String name) {
        super(DEFAULT_PIPE_BUFFER_SIZE);
        this.in = in;
        this.name = name;
        try {
            pipeOut = new PipedOutputStream(this);
        } catch (Exception e) {
            throw new IllegalStateException("FilterInputStreamViaOutputStream: error creating pipeOut: "+shortError(e));
        }
        out = instantiate(outStreamClass, pipeOut);
        thread = start();
        if (log.isDebugEnabled()) log.debug(logPrefix()+"started based on in="+in.getClass().getSimpleName());
    }

    public FilterInputStreamViaOutputStream(InputStream in, Function<OutputStream, OutputStream> outFactory) {
        this(in, outFactory, null);
    }

    public FilterInputStreamViaOutputStream(InputStream in, Function<OutputStream, OutputStream> outFactory, String name) {
        super(DEFAULT_PIPE_BUFFER_SIZE);
        this.in = in;
        this.name = name;
        try {
            pipeOut = new PipedOutputStream(this);
        } catch (Exception e) {
            throw new IllegalStateException("FilterInputStreamViaOutputStream: error creating pipeOut: "+shortError(e));
        }
        try {
            out = outFactory.apply(pipeOut);
        } catch (Exception e) {
            throw new IllegalStateException("FilterInputStreamViaOutputStream: error creating out: "+shortError(e));
        }
        thread = start();
        if (log.isDebugEnabled()) log.debug(logPrefix()+"started based on in="+in.getClass().getSimpleName());
    }

    private Thread start() {
        final Thread t = new Thread(this);
        t.setName(!empty(name) ? name : getClass().getSimpleName()+"/"+System.identityHashCode(this));
        t.setDaemon(true);
        t.start();
        return t;
    }

    private String logPrefix() { return (thread != null ? thread.getName() : !empty(name) ? name : "no-name")+": "; }

    @Override public void run() {
        try {
            final byte[] buf = new byte[DEFAULT_COPY_BUFFER_SIZE];
            int bytesRead;
            while ((bytesRead = in.read(buf)) >= 0) {
                if (log.isDebugEnabled()) log.debug(logPrefix()+"run: pumping "+bytesRead+" bytes read from in -> out");
                out.write(buf, 0, bytesRead);
            }
            out.flush();

        } catch (IOException e) {
            final String msg = "run: error copying bytes: " + shortError(e);
            log.error(msg);
            throw new RuntimeException(msg, e);

        } finally {
            closeQuietly(out);
        }
    }

    @Override public void close() {
        if (log.isDebugEnabled()) log.debug(logPrefix()+"close");
        try {
            super.close();
        } catch (Exception e) {
            log.warn("close: error closing pipeIn: "+shortError(e));
        }
        closeQuietly(in);
        closeQuietly(out);
        closeQuietly(pipeOut);
        if (thread.isAlive()) {
            background(() -> terminate(thread, THREAD_TERMINATE_TIMEOUT));
        }
    }
}
